% this script was initially created for the work "Reflection artifact
% identification in photoacoustic imaging using multi-wavelength 
% excitation" but can also be used for other work.
% It basically processes raw data acquired by the FULLPHASE probe,
% reconstructs images, segments and obtains the spectral response of 
% the segmented image features.
% Refer to the above paper for the flow chart.

% function [ Settings, Data, Objects, rekon, crossCor ] = process_data_v2(file)
% function [ Settings, Data, Objects] = process_data_v2(file, location)
% function [ Settings, Data,  ArtifactsRemovedData] = process_data_v2(file)
% function [ Settings, Data, bloodObj, Objects] = process_data_v2(file)

close all

%% read data
% [ Settings, Data ] = Read_RFe_V2c( file, 0, 0);
[ Settings, Data ] = Read_RFe_V2c( 'P:\TNW\BMPI\Users\Ho Nhu Y Nguyen\Published papers\In-plane artifact reduction\data', 0, 0);

%% preprocess data
Settings.events = unique(Settings.lm(4,:));
if Settings.events(1) == 0
    Settings.events(1) = [];
end
Settings.wavelengths = wavelength_event_converter(Settings.events);
nEvents = length(Settings.events);
signal = preprocess_data_PA(Settings, Data);

%% reconstruct images
fprintf('Reconstructing image...');
Settings.pitch = 0.3; % transducer element Settings.pitch [mm]
Settings.c =  1.5; % speed of sound mm/s
Settings.delay = 1; % time delay from laser irradiation to signal acquisition start
rekon = zeros(nEvents, Settings.halfFrameLength, Settings.finfo.nrl);
Settings.sampleDepth = Settings.c*1e6/Settings.acq.fs; % depth of each sample
Settings.acqDepth = [1, Settings.halfFrameLength]*Settings.sampleDepth;
Settings.accY = (1:64)*Settings.pitch;

maxRekon = 0;
clearestImgIdx = 1; % to find the clearest image (highest SRN)
figure
subplotRow = ceil(sqrt(nEvents));
subplotCol = ceil(nEvents/subplotRow);
for iEvent = 1:nEvents
    subplot(subplotRow, subplotCol, iEvent);
    tempRekon = rekon_OA_freqdom(squeeze(signal(iEvent,:,:)), Settings);
    rekon(iEvent,:,:) = tempRekon;
    if max(tempRekon(:)) > maxRekon
        maxRekon = max(tempRekon(:));
        clearestImgIdx = iEvent;
    end
    imagesc(Settings.accY, Settings.acqDepth, tempRekon);
    axis image
    ylabel('depth [mm]');
    xlabel('y [mm]');
    colormap(hot);
    rekon(iEvent,:,:) = tempRekon;
    title(['\lambda = ' Settings.wavelengths{iEvent} 'nm']);
end
fprintf('done!\n');

%% Segment the clearest image
% reference: https://mathworks.com/help/images/examples/detecting-a-cell-using-image-segmentation.html
fprintf('Removing reflections...');
Settings.clearestImgIdx = clearestImgIdx;
clearestImg = squeeze(rekon(clearestImgIdx, :, :)); % mean of all wavelenghts

Objects = segment_image(clearestImg, Settings);
Objects.clearestImgIdx = clearestImgIdx;
nObjs = Objects.nObjs;

for iObj = 1:nObjs
    [Objects.objPosX{iObj}, Objects.objPosY{iObj}] = ind2sub(size(clearestImg),...
        Objects.PixelIdxList{iObj});
    Objects.objWidth(iObj) = length(unique(Objects.objPosY{iObj}));
    [maxPixel, linIndex] = max(clearestImg(Objects.PixelIdxList{iObj}));
    linIndex = Objects.PixelIdxList{iObj}(linIndex);
    Objects.objPxlIdx(iObj) = linIndex;
end
figure
imagesc(Objects.objectsShow);
axis off
colormap hot;
numberedObjs = [24 26 51 54 57 58];
numberedObjs = sort(numberedObjs);

%% record the spectral response, position of each object
Objects.objPxlValueWavelength = zeros(nEvents, nObjs);
Objects.objPxlPosXWavelength = zeros(nEvents, nObjs);
Objects.objPxlPosYWavelength = zeros(nEvents, nObjs);
Objects.objPxlPosLinIndex = zeros(nObjs,1);
objMaxPixel = zeros(nObjs);
displedObjs = zeros(1, nObjs);
bloodObjMax = 0;
for iEvent = 1:nEvents
    eventImg = squeeze(rekon(iEvent, :, :));
    for iObj = 1:nObjs
        % get maximum pixel of each object
        [maxPixel, linIndex] = max(eventImg(Objects.PixelIdxList{iObj}));
        linIndex = Objects.PixelIdxList{iObj}(linIndex);
        [pxlX, pxlY] = ind2sub(size(clearestImg), linIndex);
        if iEvent == clearestImgIdx
            if bloodObjMax < maxPixel
                bloodObjMax =  maxPixel;
                bloodObj = iObj;
            end
            objMaxPixel(iObj) = maxPixel;
            Objects.objPxlPosLinIndex(iObj) = linIndex;
            text(pxlY,pxlX,num2str(iObj), 'color',[0 1 1], 'fontsize',11,...
                'fontweight','bold');
        end
        Objects.objPxlPosXWavelength(iEvent, iObj) = pxlX;
        Objects.objPxlPosYWavelength(iEvent, iObj) = pxlY;
        Objects.objPxlValueWavelength(iEvent, iObj) = maxPixel;
        Objects.objPxlMeanValueWavelength(iEvent, iObj) = mean(eventImg(Objects.PixelIdxList{iObj}));
        Objects.objSizeWavelength(iEvent, iObj) = length(Objects.PixelIdxList{iObj});
    end
end

%% group similar responses
[Objects.objPxlPosX, Objects.objPxlPosY] = ind2sub(size(clearestImg), Objects.objPxlPosLinIndex);
[crossCor pValues] = corrcoef(Objects.objPxlValueWavelength); % calculate the cross corelation coefficient
                               % of all Objects to each other
Objects.crossCor = crossCor;
Objects.pValues = pValues;
Settings.crossCorThresh = 0.95;
suspectReflInd = find(tril(crossCor) > Settings.crossCorThresh & tril(crossCor) < 1);
[suspectReflX, suspectReflY] = ind2sub(size(crossCor), suspectReflInd);
[suspectReflX, suspectReflSortedIds] = sort(suspectReflX);
suspectReflY = suspectReflY(suspectReflSortedIds);
[~, suspectReflIds] = unique(suspectReflX);

%% identify reflections
%% this part can be optimized since the angleTanThreshold is not needed
Cond.zDistanceThreshold = 1.7; % the distance threshold along depth between the
                        % absorber and the reflection to define if it is
                        % the reflection [mm]
removedObj = zeros(1, nObjs);
DevRemoveObj = struct([]); % for dev
for iGroup = 1:length(suspectReflIds)
    if removedObj(suspectReflX(suspectReflIds(iGroup))) == 0
        if iGroup == length(suspectReflIds)
            group = [suspectReflY(suspectReflIds(iGroup):end); suspectReflX(suspectReflIds(iGroup))];
        else
            group = [suspectReflY(suspectReflIds(iGroup):suspectReflIds(iGroup+1)-1); suspectReflX(suspectReflIds(iGroup))];
        end
        group(removedObj(group) > 0) = [];
        [removedObj, DevRemoveObj] = group_objects_with_conditions(group,...
            removedObj, Objects, Settings, Cond, DevRemoveObj, crossCor);
    end
end


%% remove reflections
%% DevRemoveObj can be commented to speed up, it is just  for debugging
% DevRemoveObj(arrayfun(@(x) isempty(x.abs) && isempty(x.refl), DevRemoveObj)) = [];
removedObj(removedObj == 0) = [];
Objects.objectsShowCorrected = Objects.objectsShow;
for iRemovedObj = 1:length(removedObj)
    Objects.objectsShowCorrected(Objects.PixelIdxList{removedObj(iRemovedObj)}) = 0;
end
correctedImg = deprocess_peaks_v2(Objects.objectsShowCorrected, clearestImg);

%% final steps for showing the corrected image and US image
% profile viewer

% US image
indices = Settings.lm(4,:) == 0;
frames = Data.fullFrame(indices,:,:);
tempFrame = squeeze(mean(frames));
tempFrame(Settings.cutOffIdx1+1:end-Settings.cutOffFullIdx-1, :) = detrend(tempFrame(Settings.cutOffIdx1+1:end-Settings.cutOffFullIdx-1, :), 'constant');
USFrame = rekon_US_freqdom_anglenew_compounding(tempFrame...
        ,Settings.acq.fs/1e6,0.3,Settings.c,2,0,0,5,1,1e9,0);
USFrame(278:725,:) = USFrame(253:700,:);
figure
subplot(1,3,1)
imagesc(Settings.accY, Settings.acqDepth, USFrame);
ylabel('depth [mm]');
xlabel('y [mm]');
title('Ultrasound image');
colormap hot
axis image
subplot(1,3,2)
imagesc(Settings.accY, Settings.acqDepth, clearestImg);
ylabel('depth [mm]');
xlabel('y [mm]');
title('Acquired PA image');
colormap hot
axis image
subplot(1,3,3)
imagesc(Settings.accY, Settings.acqDepth, correctedImg);
ylabel('depth [mm]');
xlabel('y [mm]');
title('Corrected PA image');
colormap hot
axis image
fprintf('done!\n');
